#include <curl/curl.h>
#include <ncurses.h>
#include <stdbool.h>
#include <unistd.h>
#include "opt_exists.c"
#include <strings.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

typedef struct {
  WINDOW *w;
  char *skip_file;
  char *to_dir;
  bool progress_enabled;
  unsigned screenline;
  int number_of_files;
  FILE *output;
  bool show_rc;
  bool copy_to_local;
} my_data;

int progress_callback(void *ptr, double dltotal, double dlnow,
    double ult, double uln)
{
  my_data *data = ptr;
  if (data->progress_enabled) {
    mvprintw(data->screenline, 38, "%6.2lf %%", dlnow/dltotal * 100);
  }
  refresh();
  return 0;
}

int start_unit(const Curl_fileinfo *finfo, void *ptr, int remains)
{
  my_data *data = ptr;
  char remains_string[15] = "";

  if (data->number_of_files == -1) {
    data->number_of_files = remains;
  }

  /* file counter */
  sprintf(remains_string, "(%d/%d)", data->number_of_files - remains + 1,
      data->number_of_files);
  mvprintw(++data->screenline, 0, "%30s %4s", finfo->filename, remains_string);

  /* -skip=file argument */
  if (data->skip_file) {
    if (strcmp(data->skip_file, finfo->filename) == 0) {
      mvprintw(data->screenline, 38, "Skipped in STARTUNIT");
      return 1;
    }
  }

  /* -to_local / -to_dir=dir argument */
  if (data->copy_to_local && finfo->filetype == CURLFILETYPE_FILE) {
    FILE *f;
    char filename[200] = ""; /* :-) */
    if (data->to_dir) {
      int ret = mkdir(data->to_dir, 0777);
      if (ret != 0) {
        if (errno != EEXIST) {
          mvprintw(data->screenline, 38, "Cannot write to %s", finfo->filename);
          return 1;
        }
      }
      strcat(filename, data->to_dir);
      strcat(filename, "/");
    }
    strcat(filename, finfo->filename);

    f = fopen(filename, "w");
    if (!f) {
      mvprintw(data->screenline, 38, "Cannot write to %s", finfo->filename);
      return 1;
    }
    else {
      data->output = f;
    }
  }
  else if (finfo->filetype == CURLFILETYPE_DIRECTORY) {
    mvprintw(data->screenline, 38, "#directory#");
  }

  /* turn on progress callback output */
  data->progress_enabled = true;

  refresh();
  return 0;
}

int end_unit(void *ptr)
{
  my_data *data = ptr;
  data->progress_enabled = false;
  if (data->output) {
    fclose(data->output);
    data->output = NULL;
  }
  return 0;
}

size_t write_data(void *buffer, size_t size, size_t nmemb, void *userp)
{
  my_data *data = userp;
  if (data->output) {
    fwrite(buffer, size, nmemb, data->output);
  }
  return size*nmemb;
}

int main(int argc, char **argv)
{
  char *optreturn;

  my_data *data = malloc(sizeof(my_data));
  bzero(data, sizeof(my_data));
  data->number_of_files = -1;
  data->screenline = -1;

  if((optreturn = opt_exists(argc, argv, "-help"))  != NULL ||
     (optreturn = opt_exists(argc, argv, "-h"))  != NULL ||
     (optreturn = opt_exists(argc, argv, "--help")) != NULL) {
    printf(
      " -help\n"
      " -show_url               (shows set url)\n"
      " -domain=ftp://user:pass@domain.tld\n"
      " -uri=/*.txt\n"
      " -repeat=N               (repeats download the same target N times)\n"
      " -show_rc                (after each curl_easy_perform() prints\n"
      "                          RETURN CODE of call)\n"
      " -skip=FILENAME          (skip file downloading in startunit callback)\n"
      " -to_local               (copies remote files to local filesystem)\n"
      " -to_dir=dirname         (target directory for -to_local)\n"
    );
    return 0;
  }

  /* init ncurses data */
  data->w = initscr();
  noecho();
  refresh();

  /* init curl */
  curl_global_init(CURL_GLOBAL_ALL);
  CURL *curl = curl_easy_init();

  char url[1000] = "";


  if((optreturn = opt_exists(argc, argv, "-domain=")) != NULL) {
    strcat(url, optreturn);
  }
  else {
    strcat(url, "ftp://curltest.howto.cz:123456@curltest.howto.cz");
  }

  if((optreturn = opt_exists(argc, argv, "-uri=")) != NULL) {
    strcat(url, optreturn);
  }

  if(opt_exists(argc, argv, "-show_url")) {
    mvprintw(++data->screenline, 0, ">> %s <<\n", url);
  }

  if (opt_exists(argc, argv, "-to_local")) {
    data->copy_to_local = true;
  }

  if ((optreturn = opt_exists(argc, argv, "-to_dir=")) != NULL) {
    data->to_dir = optreturn;
  }

  if (opt_exists(argc, argv, "-show_rc")) {
    data->show_rc = true;
  }

  if((optreturn = opt_exists(argc, argv, "-skip=")) != NULL) {
    data->skip_file = optreturn;
  }

  curl_easy_setopt(curl, CURLOPT_URL, url);
  curl_easy_setopt(curl, CURLOPT_WILDCARDMATCH, 1L);

  curl_easy_setopt(curl, CURLOPT_PROGRESSFUNCTION, progress_callback);
  curl_easy_setopt(curl, CURLOPT_NOPROGRESS, 0L);
  curl_easy_setopt(curl, CURLOPT_PROGRESSDATA, data);

  curl_easy_setopt(curl, CURLOPT_STARTUNIT, start_unit);
  curl_easy_setopt(curl, CURLOPT_ENDUNIT, end_unit);
  curl_easy_setopt(curl, CURLOPT_WRITEDATA, data);
  curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, write_data);


  /* parameter -repeat */
  int loop = 1;
  optreturn = opt_exists(argc, argv, "-repeat=");
  if (optreturn) {
    loop = atol(optreturn);
    if (loop <= 0) {
      loop = 1;
    }
  }
  for (int i = 0; i < loop; i++) {
    if (i != 0) {
      mvprintw(++data->screenline, 0,
          "==========================================================");
    }
    int rc = curl_easy_perform(curl);
    if (data->show_rc)
      mvprintw(++data->screenline, 0, "RC = %d", rc);
  }

  /* clean */
  curl_easy_cleanup(curl);
  curl_global_cleanup();

  /* sleep for some time */
  if ((optreturn = opt_exists(argc, argv, "-sleep=")) != NULL) {
    int sleep_time = atoi(optreturn);
    sleep(sleep_time);
  }

  endwin();
}
